
#include "TComDictCoder.h"
#include <string.h>

#if DUAL_CODER

CDictEncoder::CDictEncoder()
{
  m_uCurrPackSize = 0;
  m_bFirst = true;
  m_bFirstInit = true;
  m_iDictCoderLevel = 4;
  m_uDictSize = 0;
}

int CDictEncoder::ReadBlockData(void *pp, void *data, UInt *size)
{
  size_t curSize = *size;
  CSeqInStreamBuf *p = (CSeqInStreamBuf *)pp;
  if (p->rem < curSize)
    curSize = p->rem;
  memcpy(data, p->data, curSize);
  p->rem -= curSize;
  p->data += curSize;
  *size = curSize;
  return 0;
}
UChar * CDictEncoder::MatchFinder_GetPointerToCurrentPos(CMatchFinder *p) 
{
  return p->buffer;
}

UInt CDictEncoder::MatchFinder_GetNumAvailableBytes(CMatchFinder *p)
{ 
  return p->streamPos - p->pos; 
}

void CDictEncoder::MatchFinder_Init(CMatchFinder *p)
{
  UInt i;
  for (i = 0; i < p->hashSizeSum; i++)
    p->m_pHash[i] = 0;
  p->cyclicBufferPos = 0;
  p->buffer = p->bufferBase;
  p->pos = p->streamPos = p->cyclicBufferSize;
  p->result = 0;
  MatchFinder_ReadBlock(p);
  MatchFinder_SetLimits(p);
}


void CDictEncoder::MatchFinder_MoveBlock(CMatchFinder *p)
{
  memmove(p->bufferBase,
    p->buffer - p->keepSizeBefore,
    (size_t)(p->streamPos - p->pos + p->keepSizeBefore));
  p->buffer = p->bufferBase + p->keepSizeBefore;
}


int CDictEncoder::MatchFinder_NeedMove(CMatchFinder *p)
{
  return ((size_t)(p->bufferBase + p->blockSize - p->buffer) <= p->keepSizeAfter);
}

void CDictEncoder::MatchFinder_CheckAndMoveAndRead(CMatchFinder *p)
{
  if (MatchFinder_NeedMove(p))
    MatchFinder_MoveBlock(p);
  MatchFinder_ReadBlock(p);
}

void CDictEncoder::MatchFinder_ReduceOffsets(CMatchFinder *p, UInt subValue)
{
  p->posLimit -= subValue;
  p->pos -= subValue;
  p->streamPos -= subValue;
}


void CDictEncoder::MatchFinder_Normalize3(UInt subValue, UInt *items, UInt numItems)
{
  UInt i;
  for (i = 0; i < numItems; i++)
  {
    UInt value = items[i];
    if (value <= subValue)
      value = 0;
    else
      value -= subValue;
    items[i] = value;
  }
}

UInt CDictEncoder::MatchFinder_GetSubValue(CMatchFinder *p)
{
  return (p->pos - p->historySize - 1) & kNormalizeMask;
}

void CDictEncoder::MatchFinder_Normalize(CMatchFinder *p)
{
  UInt subValue = MatchFinder_GetSubValue(p);
  MatchFinder_Normalize3(subValue, p->m_pHash, p->hashSizeSum + p->numSons);
  MatchFinder_ReduceOffsets(p, subValue);
}

void CDictEncoder::MatchFinder_SetLimits(CMatchFinder *p)
{
  UInt limit = kMaxValForNormalize - p->pos;
  UInt limit2 = p->cyclicBufferSize - p->cyclicBufferPos;
  if (limit2 < limit)
    limit = limit2;
  limit2 = p->streamPos - p->pos;
  if (limit2 <= p->keepSizeAfter)
  {
    if (limit2 > 0)
      limit2 = 1;
  }
  else
    limit2 -= p->keepSizeAfter;
  if (limit2 < limit)
    limit = limit2;
  {
    UInt lenLimit = p->streamPos - p->pos;
    if (lenLimit > p->matchMaxLen)
      lenLimit = p->matchMaxLen;
    p->lenLimit = lenLimit;
  }
  p->posLimit = p->pos + limit;
}

void CDictEncoder::MatchFinder_CheckLimits(CMatchFinder *p)
{

  if (p->pos == kMaxValForNormalize)
    MatchFinder_Normalize(p);
  if (p->keepSizeAfter >= p->streamPos - p->pos)
   {
     MatchFinder_CheckAndMoveAndRead(p);
   }
  if (p->cyclicBufferPos == p->cyclicBufferSize)
    p->cyclicBufferPos = 0;
  MatchFinder_SetLimits(p);
}

UInt * CDictEncoder::Hc_GetMatchesSpec(UInt lenLimit, UInt curMatch, UInt pos, const UChar *cur, UInt *son,
  UInt _cyclicBufferPos, UInt _cyclicBufferSize, UInt cutValue,
  UInt *distances, UInt maxLen)
{
  son[_cyclicBufferPos] = curMatch;
  for (;;)
  {
    UInt delta = pos - curMatch;
    if (delta == 0)
    {	
      delta = 1;
    }
    if (cutValue-- == 0 || delta >= _cyclicBufferSize)
      return distances;
    {
      const UChar *pb = cur - delta;
      curMatch = son[_cyclicBufferPos - delta + ((delta > _cyclicBufferPos) ? _cyclicBufferSize : 0)];
      if (pb[maxLen] == cur[maxLen] && *pb == *cur)
      {
        UInt len = 0;
        while (++len != lenLimit)
          if (pb[len] != cur[len])
            break;
        if (maxLen < len)
        {
          *distances++ = maxLen = len;
          *distances++ = delta - 1;
          if (len == lenLimit)
            return distances;
        }
      }
    }
  }
}


UInt CDictEncoder::Hc3_MatchFinder_GetMatches(CMatchFinder *p, UInt *distances)
{
  UInt hash2Value, hash3Value, delta2, delta3, maxLen, offset;
  UInt lenLimit;  
  const UChar *cur;
  UInt curMatch; 
  UInt hashValue;

  UInt minLen = 3;
  UInt hashMask;

  UInt bitsHash3;
  switch(p->level)
  {
  case 1:
    bitsHash3 = 8;
    break;
  case 2:
    bitsHash3 = 9;
    break;
  case 3: 
    bitsHash3 = 10;
    break;
  case 4:
    bitsHash3 = 12;
    break;
  case 5:
    bitsHash3 = 14;
    break;
  default:
    bitsHash3 = 16;
    break;
  }
  
  hashMask = (1<<bitsHash3)-1;



  lenLimit = p->lenLimit; 
  {
    if (lenLimit < minLen) 
    {
      ++p->cyclicBufferPos; 
      p->buffer++; 
      if (++p->pos == p->posLimit) 
        MatchFinder_CheckLimits(p);
      return 0; 
    }
  }
  cur = p->buffer;

  { 
    UInt temp = p->crc[cur[0]] ^ cur[1]; 
    hash2Value = temp & (kHash2Size - 1); 
    hash3Value = (temp ^ ((UInt)cur[2] << 8)) & hashMask; 
    hashValue = hash3Value;
  }


  delta2 = p->pos - p->m_pHash[                hash2Value];
  delta3 = p->pos - p->m_pHash[kHash2Size + hash3Value];
  curMatch = p->m_pHash[kHash2Size+hash3Value];



  p->m_pHash[                hash2Value] =
    p->m_pHash[kHash2Size + hash3Value] =p->pos;
  

  //it's not necessary for search when the source is rec data of HM
  if (!p->m_bEncOrg)
  {
    offset = 0;

    ++p->cyclicBufferPos; 
    p->buffer++; 
    if (++p->pos == p->posLimit) 
      MatchFinder_CheckLimits(p);
    return offset;
  }

  if (delta2 == 0)
  {
    delta2 = 1;
  }
  if (delta3 == 0)
  {
    delta3 = 1;
  }

  maxLen = 1;
  offset = 0;
  if (delta2 < p->cyclicBufferSize && *(cur - delta2) == *cur)
  {
    distances[0] = maxLen = 1;
    distances[1] = delta2 - 1;
    offset = 2;
  }
  if (delta2 != delta3 && delta3 < p->cyclicBufferSize && *(cur - delta3) == *cur)
  {
 
    maxLen = 1;

    distances[offset + 1] = delta3 - 1;
    offset += 2;
    delta2 = delta3;
  }
  if (offset != 0)
  {
    for (; maxLen != lenLimit; maxLen++)
      if (cur[(ptrdiff_t)maxLen - delta2] != cur[maxLen])
        break;
    distances[offset - 2] = maxLen;
    if (maxLen == lenLimit)
    {
      p->m_pSon[p->cyclicBufferPos] = curMatch;
      ++p->cyclicBufferPos; 
      p->buffer++; 
      if (++p->pos == p->posLimit) 
        MatchFinder_CheckLimits(p);
      return offset;
    }
  }
  if (maxLen < 3)
    maxLen = 3;
  offset = (UInt)(Hc_GetMatchesSpec(lenLimit, curMatch, p->pos, p->buffer, p->m_pSon, p->cyclicBufferPos, p->cyclicBufferSize, p->m_uCutValue,distances + offset, maxLen) - (distances));

  ++p->cyclicBufferPos; 
  p->buffer++; 
  if (++p->pos == p->posLimit) 
    MatchFinder_CheckLimits(p);		 
  return offset;
}

void CDictEncoder::MatchFinder_ReadBlock(CMatchFinder *p)
{
  for (;;)
  {
    UChar *dest = p->buffer + (p->streamPos - p->pos);
    size_t size = (p->bufferBase + p->blockSize - dest);
    if (size == 0)
      return;
    //p->result = ReadBlockData(p->stream, dest, &size);
    p->result = ReadBlockData((void *)p->stream, (void *)dest, (UInt *)(&size));
    if (p->result != 0)
      return;
    if (size == 0)
    {
      return;
    }
    p->streamPos += (UInt)size;
    if (p->streamPos - p->pos > p->keepSizeAfter)
      return;
  }
}

void CDictEncoder::MatchFinder_ReadIfRequired(CMatchFinder *p)
{
  if (p->keepSizeAfter >= p->streamPos - p->pos)
    MatchFinder_ReadBlock(p);

  if (p->pos == p->posLimit) 
    MatchFinder_CheckLimits(p);
}


UInt CDictEncoder::ReadMatchDistances(UInt *numDistancePairsRes)
{
  UInt lenRes = 0, numPairs;
  m_uNumAvail = MatchFinder_GetNumAvailableBytes(&m_matchFinderBase);
  numPairs = Hc3_MatchFinder_GetMatches(&m_matchFinderBase, m_uMatches); //numPairs  len ?

  if (numPairs > 0)
  {
    lenRes = m_uMatches[numPairs - 2];
    if (lenRes == m_uNumFastBytes)
    {
      const UChar *pby = MatchFinder_GetPointerToCurrentPos(&m_matchFinderBase) - 1;
      UInt distance = m_uMatches[numPairs - 1] + 1;

      UInt numAvail = m_uNumAvail;

      if (distance == 0)
      {
        distance = 1;
      }
      if (numAvail > MATCH_LEN_MAX)
        numAvail = MATCH_LEN_MAX;
      {
        const UChar *pby2 = pby - distance;
        lenRes = 0;
        for (; lenRes < numAvail && pby[lenRes] == pby2[lenRes]; lenRes++);
      }
    }
  }
  m_uAdditionalOffset++;
  *numDistancePairsRes = numPairs;
  return lenRes;
}

void CDictEncoder::WriteBits(UInt & uiDst,UInt value, UInt numBits, UInt uiStartPos)
{
  value=value<<(32-uiStartPos-numBits+1);
  uiDst=uiDst|value;
}
void CDictEncoder::WriteStrMapResult(int len, int pos,const UChar * pCurBuf)
{
  UInt iDst = 0;
  UInt c = 0;
  //   UInt tmp,rLen,rPos;
  int numCoded = pCurBuf-m_matchFinderBase.bufferBase;
  int litFlag=0;
  if (len == 1 && pos == (UInt)-1)
  {
    if (numCoded>=3)
    {
      litFlag = (pCurBuf[0]-pCurBuf[-3])<0;
      c = (UInt)ABS_DIFF(pCurBuf[0],pCurBuf[-3]);
    }
    else
      c = (UInt)pCurBuf[0];

    WriteBits(iDst,0,1,1);
    WriteBits(iDst,litFlag,1,2);
    WriteBits(iDst,c,8,3);
  }
  else
  {
    WriteBits(iDst,1,1,1);
    WriteBits(iDst,len,9,2);
    WriteBits(iDst,pos,22,11);
  }
  m_pStrMapBufList.push_back(iDst);
  m_uCurrPackSize++;
}


void CDictEncoder::Hc3_MatchFinder_Skip(CMatchFinder *p, UInt num)
{
  UInt hashMask;
  UInt bitsHash3;
  switch(p->level)
  {
  case 1:
    bitsHash3 = 8;
    break;
  case 2:
    bitsHash3 = 9;
    break;
  case 3: 
    bitsHash3 = 10;
    break;
  case 4:
    bitsHash3 = 12;
    break;
  case 5:
    bitsHash3 = 14;
    break;
  default:
    bitsHash3 = 16;
    break;
  }

  hashMask = (1<<bitsHash3)-1;


  do
  {
    UInt hash2Value, hash3Value;

    UInt lenLimit,hashValue,curMatch; 
    const UChar *cur; 
    lenLimit = p->lenLimit; 
    { 
      if (lenLimit < 3) 
      {
        ++p->cyclicBufferPos; 
        p->buffer++; 
        if (++p->pos == p->posLimit) 
          MatchFinder_CheckLimits(p);

        continue; 
      }
    } 
    cur = p->buffer;


    UInt temp = p->crc[cur[0]] ^ cur[1]; 
    hash2Value = temp & (kHash2Size - 1); 
    hash3Value = (temp ^ ((UInt)cur[2] << 8)) & hashMask; 
    hashValue = hash3Value;

    curMatch = p->m_pHash[kHash2Size + hashValue];

    p->m_pHash[                hash2Value] =
      p->m_pHash[kHash2Size + hash3Value] = p->pos;
    p->m_pSon[p->cyclicBufferPos] = curMatch;

    ++p->cyclicBufferPos; 
    p->buffer++; 
    if (++p->pos == p->posLimit) 
      MatchFinder_CheckLimits(p);
  }
  while (--num != 0);
}


void CDictEncoder::MovePos(UInt num)
{
  if (num != 0)
  {
    m_uAdditionalOffset += num;
    Hc3_MatchFinder_Skip(&m_matchFinderBase, num);
  }
}

UInt CDictEncoder::GetOptimumFast(UInt *backRes)
{
  UInt numAvail, mainLen, mainDist, numPairs, repIndex, repLen, i;
  const UChar *data;
  const UInt *matches;

  if (m_uAdditionalOffset == 0)
    mainLen = ReadMatchDistances(&numPairs);
  else
  {
    mainLen = m_uLongestMatchLength;
    numPairs = m_uNumPairs;
  }

  numAvail = m_uNumAvail;
  *backRes = (UInt)-1;

  if (!m_matchFinderBase.m_bEncOrg)
  {
    return 1;
  }

  if (numAvail < 2)
    return 1;
  if (numAvail > MATCH_LEN_MAX)
    numAvail = MATCH_LEN_MAX;
  data = MatchFinder_GetPointerToCurrentPos(&m_matchFinderBase) - 1;

  repLen = repIndex = 0;
  for (i = 0; i < NUM_REPS; i++)
  {
    UInt len;
    const UChar *data2 = data - (m_uReps[i] + 1);
    if (data[0] != data2[0] || data[1] != data2[1])
      continue;
    for (len = 2; len < numAvail && data[len] == data2[len]; len++);

    if (len >= m_uNumFastBytes)
    {
      *backRes = i;
      MovePos(len - 1);
      return len;
    }
    if (len > repLen)
    {
      repIndex = i;
      repLen = len;
    }
  }

  matches = m_uMatches;
  if (mainLen >= m_uNumFastBytes)
  {
    *backRes = matches[numPairs - 1] + NUM_REPS;
    MovePos(mainLen - 1);
    return mainLen;
  }

  mainDist = 0; 
  if (mainLen >= 2)
  {
    mainDist = matches[numPairs - 1];
    while (numPairs > 2 && mainLen == matches[numPairs - 4] + 1)
    {
      if (!ChangePair(matches[numPairs - 3], mainDist))
        break;
      numPairs -= 2;
      mainLen = matches[numPairs - 2];
      mainDist = matches[numPairs - 1];
    }
    if (mainLen == 2 && mainDist >= 0x80)
      mainLen = 1;
  }

  if (repLen >= 2 && (
    (repLen + 1 >= mainLen) ||
    (repLen + 2 >= mainLen && mainDist >= (1 << 9)) ||
    (repLen + 3 >= mainLen && mainDist >= (1 << 15))))
  {
    *backRes = repIndex;
    MovePos(repLen - 1);
    return repLen;
  }

  if (mainLen < 2 || numAvail <= 2)
    return 1;

  m_uLongestMatchLength = ReadMatchDistances(&m_uNumPairs);
  if (m_uLongestMatchLength >= 2)
  {
    UInt newDistance = matches[m_uNumPairs - 1];
    if ((m_uLongestMatchLength >= mainLen && newDistance < mainDist) ||
      (m_uLongestMatchLength == mainLen + 1 && !ChangePair(mainDist, newDistance)) ||
      (m_uLongestMatchLength > mainLen + 1) ||
      (m_uLongestMatchLength + 1 >= mainLen && mainLen >= 3 && ChangePair(newDistance, mainDist)))
      return 1;
  }

  data = MatchFinder_GetPointerToCurrentPos(&m_matchFinderBase) - 1;
  for (i = 0; i < NUM_REPS; i++)
  {
    UInt len, limit;
    const UChar *data2 = data - (m_uReps[i] + 1);

    if (data[0] != data2[0] || data[1] != data2[1])
      continue;
    limit = mainLen - 1;

    for (len = 2; len < limit && data[len] == data2[len]; len++);
    if (len >= limit)
      return 1;
  }
  *backRes = mainDist + NUM_REPS;
  MovePos(mainLen - 2);
  return mainLen;
}

void CDictEncoder::MatchFinder_Construct(CMatchFinder *p)
{
  UInt i;
  p->bufferBase = 0;
  p->m_pHash = 0;

  for (i = 0; i < 256; i++)
  {
    UInt r = i;
    int j;
    for (j = 0; j < 8; j++)
      r = (r >> 1) ^ (K_CRC_POLY & ~((r & 1) - 1));
    p->crc[i] = r;
  }
}

void CDictEncoder::MatchFinder_Free(CMatchFinder *p)
{
  if(p->bufferBase)
  {
    delete[] p->bufferBase;
    p->bufferBase = NULL;
  }
  if (p->m_pHash)
  {
    delete[] p->m_pHash;
    p->m_pHash = NULL;
  }
}

int CDictEncoder::DictInWindow_Create(CMatchFinder *p, UInt keepSizeReserv)
{
  UInt blockSize = p->keepSizeBefore + p->keepSizeAfter + keepSizeReserv;

  if (p->bufferBase == 0 || p->blockSize != blockSize)
  {
    p->blockSize = blockSize;
    p->bufferBase = new UChar[blockSize*sizeof(UChar)];
  }
  return (p->bufferBase != 0);
}

int CDictEncoder::MatchFinder_Create(CMatchFinder *p, UInt historySize, UInt matchMaxLen, UInt keepAddBufferAfter)
{
  UInt sizeReserv;

  sizeReserv = m_uCuLen*2;

  p->keepSizeBefore = historySize+1;

  p->keepSizeAfter = keepAddBufferAfter;


  /* we need one additional byte, since we use MoveBlock after pos++ and before dictionary using */
  if (DictInWindow_Create(p, sizeReserv))
  {
    UInt newCyclicBufferSize;
    UInt prevSize = p->hashSizeSum + p->numSons;
    UInt newSize;
    UInt bitsHash3;  

    p->historySize = historySize;


    newCyclicBufferSize = (historySize /* >> p->skipModeBits */) + 1;

    p->matchMaxLen = matchMaxLen;

    switch(p->level)
    {
    case 1:
      bitsHash3 = 8;
      break;
    case 2:
      bitsHash3 = 9;
      break;
    case 3: 
      bitsHash3 = 10;
      break;
    case 4:
      bitsHash3 = 12;
      break;
    case 5:
      bitsHash3 = 14;
      break;
    default:
      bitsHash3 = 16;
      break;
    }


    p->hashSizeSum = (1<<bitsHash3)+kHash2Size;


    p->cyclicBufferSize = newCyclicBufferSize;
    p->numSons = newCyclicBufferSize;
    newSize = p->hashSizeSum + p->numSons;
    if (p->m_pHash != 0 && prevSize == newSize)
      return 1;

    p->m_pHash = new UInt[newSize*sizeof(UInt)];
    if (p->m_pHash != 0)
    {
      p->m_pSon = p->m_pHash + p->hashSizeSum;
      return 1; 
    }
  }
  MatchFinder_Free(p);
  return 0;
}

void CDictEncoder::DictEnc_FastPosInit()
{
  int c = 2, slotFast;
  int kNumLogBits = 11;
  m_pbFastPos[0] = 0;
  m_pbFastPos[1] = 1;


  for (slotFast = 2; slotFast < kNumLogBits*2; slotFast++)
  {
    UInt k = (1 << ((slotFast >> 1) - 1));
    UInt j;
    for (j = 0; j < k; j++, c++)
      m_pbFastPos[c] = (UChar)slotFast;
  } 
}

void CDictEncoder::DictEnc_Construct()
{
  MatchFinder_Construct(&m_matchFinderBase);
  DictEnc_FastPosInit();
}


int CDictEncoder::DictEnc_Alloc()
{

  if (!MatchFinder_Create(&m_matchFinderBase, m_uDictSize, m_uNumFastBytes, MATCH_LEN_MAX))
      return 1;
 
  return 0;
}

int CDictEncoder::DictEnc_AllocAndInit()
{
  UInt i,result;
  result = DictEnc_Alloc();
  if(result)
  {
    return result;
  }
  for (i = 0 ; i < NUM_REPS; i++)
    m_uReps[i] = 0;
  m_uAdditionalOffset = 0;

  m_uNowPos64 = 0;
  return 0;
} 

void CDictEncoder::Init(UChar * pSrcBuf)
{

  if (m_bFirstInit)
  {
    DictEnc_Construct();
    m_bFirstInit = false;
  }

  int level = m_iDictCoderLevel;

  if (m_uDictSize==0)
  {
    if (level < 1 || level >7)
    {
      printf("level must between 1~7\n");
      return;
    }
    else if (level == 1)
    {
      m_uDictSize = (1<<12);
    }
    else if (level == 2)
    {
      m_uDictSize = (1<<13);
    }
    else if (level > 2)
    {
      m_uDictSize = 1 << (level * 2 + 8);
    }
  } 
  m_uDictSize -= 2;

  m_uNumFastBytes = MATCH_LEN_MAX;

  m_matchFinderBase.m_uCutValue = MATCH_FINDER_CUT_VALUE;
 
  m_cSeqBufInStream.data = pSrcBuf;

  m_uPrevPackSize = 0;
  m_bFirst = true;

  m_matchFinderBase.level = level;
    
  DictEnc_AllocAndInit();
   
}

void CDictEncoder::DictCodeOneBlock()
{
   UInt nowPos32, startPos32;
   const UChar * pCurBuf;
   int i;

   if (m_bFirst)
  { 
    m_matchFinderBase.stream = &m_cSeqBufInStream;
    MatchFinder_Init(&m_matchFinderBase);
    m_bFirst = false;
  }

  MatchFinder_ReadIfRequired(&m_matchFinderBase);

  nowPos32 = (UInt)m_uNowPos64;
  startPos32 = nowPos32;

  m_uCurrPackSize = 0;

  if (m_uNowPos64 == 0)
  {
    UInt numPairs;
    if (MatchFinder_GetNumAvailableBytes(&m_matchFinderBase) == 0)
          return ;
    ReadMatchDistances(&numPairs);
    m_uAdditionalOffset--;
    nowPos32++;
    if (m_matchFinderBase.m_bEncOrg)
    {
      WriteStrMapResult(1,-1,m_matchFinderBase.bufferBase);
    }
  }

  if (MatchFinder_GetNumAvailableBytes(&m_matchFinderBase) != 0)
    for (;;)
    {
      UInt pos, len;

      //if last cu roll back(coded by HM),then reset m_uReps[] to zero, otherwise decode error will happen.
      if (m_matchFinderBase.m_bEncOrg && m_bIsLastCuRollBack)
      {
        for (i = 0; i< NUM_REPS; i++)
        {
          m_uReps[i] = 0;
        }
       m_bIsLastCuRollBack = 0;
      }

      len = GetOptimumFast(&pos);

      if (!m_matchFinderBase.m_bEncOrg)
      {

        m_uAdditionalOffset -= len;

        if ((m_uAdditionalOffset == 0) && (MatchFinder_GetNumAvailableBytes(&m_matchFinderBase) == 0))
        {
          m_bIsLastCuRollBack = 1;
          break;
        }
        continue;
      }


      if (m_matchFinderBase.m_bEncOrg)
      {

        pCurBuf = MatchFinder_GetPointerToCurrentPos(&m_matchFinderBase) - m_uAdditionalOffset;

        WriteStrMapResult(len,pos,pCurBuf);

 				if(len == 1 && pos == (UInt)-1)
         {  }
        else if (pos < NUM_REPS)
        {
          int dis = m_uReps[pos];
          for (i=pos;i>0;i--)
          {
            m_uReps[i] = m_uReps[i-1];
          }
          m_uReps[0] = dis;
        }
        else
        {
          int posEnc = pos - NUM_REPS;
          for (i=NUM_REPS-1;i>0;i--)
          {
            m_uReps[i] = m_uReps[i-1];
          }
          m_uReps[0] = posEnc;
        }
        m_uAdditionalOffset -= len;
        nowPos32 += len;

        if ((m_uAdditionalOffset == 0) && (MatchFinder_GetNumAvailableBytes(&m_matchFinderBase) == 0))
        {
          break;
        }
        continue;
      }
    }

    m_uNowPos64 += nowPos32 - startPos32;
    return ;
}
TComDictEnc::TComDictEnc()
{
  m_pDictEncoder = NULL;
  m_pCuBuf = NULL;
  m_pPackSizeBuf = NULL;
  m_bNeedRollBack = false; 
  m_bIsFirstRollBack = true;
  m_iDictCoderLevel = 4;
  m_uiStartPackPos = 0;
}

void TComDictEnc::DictEnc_SaveState(void * pp)
{
  CDictEncoder *p = (CDictEncoder *)pp;
  CSaveState *dest = &p->m_cSaveState;

  dest->mf = *(p->GetMatchFinderObj());
  dest->m_uNowPos64 = p->m_uNowPos64;

  memcpy(dest->m_uReps, p->m_uReps, sizeof(p->m_uReps));

} 

void TComDictEnc::DictEnc_RestoreState(void * pp)
{
	UChar * bufferBak;
	CDictEncoder *dest = (CDictEncoder *)pp;
	const CSaveState *p = &dest->m_cSaveState;


	bufferBak = dest->GetMatchFinderObj()->buffer; 
	dest->SetMatchFinderObj(p->mf);
	dest->m_uNowPos64 = p->m_uNowPos64;

	if (bufferBak < p->mf.buffer)
	{
		dest->GetMatchFinderObj()->buffer = bufferBak - m_uCuLen;
	}
  memcpy(dest->m_uReps, p->m_uReps, sizeof(p->m_uReps));
}

void TComDictEnc::InitEncoder(TComSlice* pcSlice,UInt cuCnt)
{
  UInt frame_size;
  static int isFirst = 1;

  if (isFirst || m_bNeedResetEncoder)
  {
    isFirst = 0;
    m_bNeedResetEncoder = 0;
  }
  else 
  {
    return;
  }

  m_uCuWidth = pcSlice->getSPS()->getMaxCUWidth();
  m_uCuHeight = pcSlice->getSPS()->getMaxCUHeight();

  m_uCuLen = m_uCuWidth * m_uCuHeight * 3;


  m_uPicWidth = pcSlice->getPic()->getPicYuvOrg()->getWidth(COMPONENT_Y);
  m_uPicHeight = pcSlice->getPic()->getPicYuvOrg()->getHeight(COMPONENT_Y);

  frame_size = m_uPicWidth*m_uPicHeight;
  m_uPicLen = frame_size*3;

  m_uPicLenPad = cuCnt * m_uCuLen;
  //input cu buffer
  if (m_pCuBuf == NULL)
  {
      m_pCuBuf = new UChar[m_uCuLen*sizeof(UChar)];
  }

  if (m_pPackSizeBuf == NULL)
  {
    m_pPackSizeBuf = new UInt[cuCnt*sizeof(UInt)];
  }
  memset(m_pPackSizeBuf,0,cuCnt*sizeof(int));

  if (m_pDictEncoder == NULL)
  {
      m_pDictEncoder = new CDictEncoder;
      m_pDictEncoder->m_uCuLen = m_uCuLen;
      m_pDictEncoder->m_iDictCoderLevel = m_iDictCoderLevel;
  }
  m_pDictEncoder->Init(m_pCuBuf);
    
   m_uCuEncNumByLz = 0;
}


void TComDictEnc::TranposeCu(UChar * pCuBuf)
{
  int wdt,hgt,i,j;
  UChar *pY,*pU,*pV;
  UChar *pTmp,*tY,*tU,*tV;
  UInt cuSizeY = m_uCuWidth * m_uCuHeight;
  pTmp = new UChar[m_uCuLen*sizeof(UChar)];
  memcpy(pTmp,pCuBuf,m_uCuLen*sizeof(UChar));

  pY = pCuBuf;
  pU = pY + cuSizeY;
  pV = pU + cuSizeY;

  wdt = hgt = m_uCuWidth;

  for (i=0;i<wdt;i++)
  {
    tY = pTmp + i;
    for (j=0;j<hgt;j++)
    {
      *pY++ = *tY;
      tY += wdt;
    }
  }

  for (i=0;i<wdt;i++)
  {
    tU = pTmp + cuSizeY + i;
    for (j=0;j<hgt;j++)
    {
      *pU++ = *tU;
      tU += wdt;
    } 
  }
  for (i=0;i<wdt;i++)
  {
    tV = pTmp + cuSizeY*2  + i;
    for (j=0;j<hgt;j++)
    {
      *pV++ = *tV;
      tV += wdt;
    }
  }
  if (pTmp)
  {
    delete[] pTmp;
    pTmp = NULL;
  }
}
void TComDictEnc::ReOrderDataEncYuv444(UChar * pCuBuf)
{
  int i;
  UChar * pTmp;
  UChar * pY, *tY, *tU, *tV;
  pTmp = new UChar[m_uCuLen*sizeof(UChar)];
  memcpy(pTmp,pCuBuf,m_uCuLen*sizeof(UChar));
  tY = pTmp;
  tU = tY + m_uCuWidth * m_uCuHeight;
  tV = tU + m_uCuWidth * m_uCuHeight;
  
  pY = pCuBuf;

  for (i = 0; i< m_uCuWidth*m_uCuHeight;i++)
  {
    *pY++ = *tY++;
    *pY++ = *tU++;
    *pY++ = *tV++;
  }

  if(pTmp)
  {
    delete[] pTmp;
    pTmp = NULL;
  }
}


void TComDictEnc::CopyCuFromPicYuv444(TComPic* pcPic,UChar* pDst ,UInt cuIdx,bool bEncOrg)
{
  int  i,j;
  int strideY, strideUV;
  UChar *pDstBase;

  Pel* pSrc;

  UInt cuYHeight = m_uCuHeight;
  UInt cuYWidth = m_uCuWidth;
  UInt cuUVHeight, cuUVWidth;
  //static UInt cuNumEncByLz = 0;

  UInt picWidth, picHeight, cuNumInWidth, cuNumInHeight;
  Double dBorderCuWidthScale,dBorderCuHeightScale;
  bool bBorderCuX = false,bBorderCuY = false;

  picWidth = pcPic->getPicYuvOrg()->getWidth(COMPONENT_Y);
  picHeight = pcPic->getPicYuvOrg()->getHeight(COMPONENT_Y);

  dBorderCuWidthScale = (Double)picWidth/(Double)m_uCuWidth - picWidth / m_uCuWidth;
  dBorderCuHeightScale  = (Double)picHeight / (Double)m_uCuHeight - picHeight/m_uCuHeight;

  cuNumInWidth = pcPic->getFrameWidthInCU();
  cuNumInHeight = pcPic->getFrameHeightInCU();

  bBorderCuX = (cuIdx % cuNumInWidth == (cuNumInWidth - 1)) ;
  bBorderCuY = (cuIdx / cuNumInWidth == (cuNumInHeight - 1));

  pDstBase = pDst;

  cuUVWidth = cuYWidth;
  cuUVHeight = cuYHeight;

  if (bEncOrg)
  {
    pSrc= pcPic->getPicYuvOrg()->getAddr(COMPONENT_Y,cuIdx);
  }
  else
  {
    pSrc= pcPic->getPicYuvRec()->getAddr(COMPONENT_Y,cuIdx);

  }
  strideUV = strideY = pcPic->getPicYuvOrg()->getStride(COMPONENT_Y);

  for (i=0;i<cuYHeight;i++)
  {
    for (j=0;j<cuYWidth;j++)
    {
      *pDst++ = *(pSrc+j);
    }
    pSrc+=strideY;
  }
  if (bBorderCuX && dBorderCuWidthScale)
  {
    int padW = (int)(m_uCuWidth - m_uCuWidth*dBorderCuWidthScale);
    UChar * pTmp = pDst - cuYHeight*cuYWidth;
    for (i = 0 ;i < cuYHeight; i++)
    {
      memset(pTmp+(m_uCuWidth-padW),0,padW);
      pTmp += m_uCuWidth;
    }
    pTmp = NULL;
  }
  if (bBorderCuY && dBorderCuHeightScale)
  {
    int padH =(int)( m_uCuHeight - m_uCuHeight*dBorderCuHeightScale);
    UChar * pTmp = pDst - padH * m_uCuWidth;
    for (i=0;i<padH;i++)
    {
      memset(pTmp,0,m_uCuWidth);
      pTmp += m_uCuWidth;
    }
    pTmp = NULL;
  }

  //copy U data
  if (bEncOrg)
  {
    pSrc = pcPic->getPicYuvOrg()->getAddr(COMPONENT_Cb,cuIdx);	
  }
  else
  {
    pSrc = pcPic->getPicYuvRec()->getAddr(COMPONENT_Cb,cuIdx);	
  }

  for (i=0;i<cuUVHeight;i++)
  {
    for (j=0;j<cuUVWidth;j++)
    {
      *pDst++ = *(pSrc+j);
    }
    pSrc+=strideUV;
  }
  if (bBorderCuX && dBorderCuWidthScale)
  {
    int padW =(int)( m_uCuWidth - m_uCuWidth*dBorderCuWidthScale);
    UChar * pTmp = pDst - cuYHeight*cuYWidth;
    for (i = 0 ;i < cuUVHeight; i++)
    {
      memset(pTmp+(m_uCuWidth-padW),0,padW);
      pTmp += m_uCuWidth;
    }
    pTmp = NULL;
  }
  if (bBorderCuY && dBorderCuHeightScale)
  {
    int padH =(int)( m_uCuHeight - m_uCuHeight*dBorderCuHeightScale);
    UChar * pTmp = pDst - padH * m_uCuWidth;
    for (i=0;i<padH;i++)
    {
      memset(pTmp,0,m_uCuWidth);
      pTmp += m_uCuWidth;
    }
    pTmp = NULL;
  }

  //copy V data
  if (bEncOrg)
  {
    pSrc = pcPic->getPicYuvOrg()->getAddr(COMPONENT_Cr,cuIdx);	
  }
  else
  {
    pSrc = pcPic->getPicYuvRec()->getAddr(COMPONENT_Cr,cuIdx);	
  }

  for (i=0;i<cuUVHeight;i++)
  {
    for (j=0;j<cuUVWidth;j++)
    {
      *pDst++ = *(pSrc+j);
    } 
    pSrc+=strideUV;
  } 

  if (bBorderCuX && dBorderCuWidthScale)
  {
    int padW =(int)( m_uCuWidth - m_uCuWidth*dBorderCuWidthScale);
    UChar * pTmp = pDst - cuYHeight*cuYWidth;
    for (i = 0 ;i < cuUVHeight; i++)
    {
      memset(pTmp+(m_uCuWidth-padW),0,padW);
      pTmp += m_uCuWidth;
    }
    pTmp = NULL;
  }
  if (bBorderCuY && dBorderCuHeightScale)
  {
    int padH =(int)( m_uCuHeight - m_uCuHeight*dBorderCuHeightScale);
    UChar * pTmp = pDst - padH * m_uCuWidth;
    for (i=0;i<padH;i++)
    {
      memset(pTmp,0,m_uCuWidth);
      pTmp += m_uCuWidth;
    }
    pTmp = NULL;
  }

  //reorder data
#if REORDER_DATA
  TranposeCu(m_pCuBuf);
  ReOrderDataEncYuv444(m_pCuBuf);
#endif

  m_pDictEncoder->m_cSeqBufInStream.data = m_pCuBuf;
  m_pDictEncoder->m_cSeqBufInStream.rem = m_uCuLen;

}

void TComDictEnc::WriteCu2RecYuv444(TComDataCU* pcCU, UInt cuIdx) 
{  
  int i,j, cuSize,chromaCuSize;
  int strideY, strideUV;;
  Pel  *pDstY, *pDstU, *pDstV;
  Pel * pSrcY,*pSrcU,*pSrcV;

  cuSize = m_uCuLen;

  pDstY = pcCU->getPic()->getPicYuvRec()->getAddr(COMPONENT_Y,cuIdx);
  pDstU = pcCU->getPic()->getPicYuvRec()->getAddr(COMPONENT_Cb,cuIdx);
  pDstV = pcCU->getPic()->getPicYuvRec()->getAddr(COMPONENT_Cr,cuIdx);

  pSrcY =	pcCU->getPic()->getPicYuvOrg()->getAddr(COMPONENT_Y,cuIdx);
  pSrcU = pcCU->getPic()->getPicYuvOrg()->getAddr(COMPONENT_Cb,cuIdx);
  pSrcV = pcCU->getPic()->getPicYuvOrg()->getAddr(COMPONENT_Cr,cuIdx);

  strideY = pcCU->getPic()->getPicYuvRec()->getStride(COMPONENT_Y);

  strideUV = pcCU->getPic()->getPicYuvRec()->getStride(COMPONENT_Y);
  chromaCuSize = m_uCuWidth;

  //copy Y
  for (i=0;i<m_uCuWidth;i++)
  {
    for (j=0;j<m_uCuHeight;j++)
    {
      *pDstY++ = *pSrcY++;
    }
    pDstY += strideY - m_uCuWidth;
    pSrcY += strideY - m_uCuHeight;
  }

  //copy U
  for (i=0;i<chromaCuSize;i++)
  {
    for(j=0;j<chromaCuSize;j++)
    {
      *pDstU++ = *pSrcU++;
    }
    pDstU += strideUV - (chromaCuSize);
    pSrcU += strideUV - (chromaCuSize);
  }

  //copy V
  for (i=0;i<chromaCuSize;i++)
  {
    for(j=0;j<chromaCuSize;j++)
    {
      *pDstV++ = *pSrcV++;
    }
    pDstV += strideUV - (chromaCuSize);
    pSrcV += strideUV - (chromaCuSize);
  }
}

void TComDictEnc::destroy()
{
  if (m_pCuBuf)
  {
    delete[] m_pCuBuf;
    m_pCuBuf = NULL;
  }
  if(m_pPackSizeBuf)
  {
    delete[] m_pPackSizeBuf;
    m_pPackSizeBuf = NULL;
  }

  if (m_pDictEncoder)
  {
     m_pDictEncoder->MatchFinder_Free(&m_pDictEncoder->m_matchFinderBase);

    delete m_pDictEncoder;
    m_pDictEncoder = NULL;
  }
}

int CDictEncoder::GetPosSlotFunc(int pos)
{
  int posSlot;
  int kNumLogBits = 11;

  if (pos < 128)
    posSlot = m_pbFastPos[pos]; 
  else 
  {
    UInt i = 6 + ((kNumLogBits - 1) & (0 - (((((UInt)1 << (kNumLogBits + 6)) - 1) - pos) >> 31))); 
    posSlot = m_pbFastPos[pos >> i] + (i * 2); 
  }

  return posSlot;

}

void TComDictEnc::WriteBitstreamDictEnc( UInt cuIdx, UInt cuEndIdx, TEncEntropy * pEncEty, UInt isLastCuRollBack)
{
   pEncEty->m_pcEntropyCoderIf->codeDictEnc(m_uiStartPackPos, m_uCuLen, m_pDictEncoder,cuIdx==cuEndIdx,isLastCuRollBack); 
   m_uiStartPackPos += m_pPackSizeBuf[cuIdx];
}

void TComDictEnc::DictCompressCu(UInt cuIdx, UInt cuEndIdx, bool bEncOrg, TComSlice* pcSlice)
{
  int idx,i;
  //int remained = 0;
  UInt frame_size;
  TComPic * pcPic = NULL;

  frame_size = m_uPicWidth * m_uPicHeight;

  pcPic = pcSlice->getPic();

  if (m_bNeedRollBack)
  {
    if (bEncOrg)
    {
      idx = cuIdx;
    }
    else
    {
      idx = cuIdx+1;
    }
    DictEnc_RestoreState(m_pDictEncoder);

    m_pDictEncoder->m_uPrevPackSize -= m_pDictEncoder->m_uCurrPackSize;
    m_pPackSizeBuf[idx-1] = 0;

    for(i = 0 ;i < m_pDictEncoder->m_uCurrPackSize;i++)
    {
      m_pDictEncoder->m_pStrMapBufList.pop_back();
    }
    m_pDictEncoder->m_uCurrPackSize = 0;

  }
  if (m_pDictEncoder->m_uPrevPackSize)
  {
    m_pDictEncoder->m_bFirst = false;
  }	
  else if(m_bNeedRollBack)
  {
    if(m_bIsFirstRollBack)
    {
      m_bNeedResetEncoder = 1;
      InitEncoder(pcSlice, cuEndIdx+1);
      m_bIsFirstRollBack = false;
     }
  } 
  DictEnc_SaveState(m_pDictEncoder);

  CopyCuFromPicYuv444(pcPic,m_pCuBuf, cuIdx,bEncOrg);

  m_pDictEncoder->m_matchFinderBase.m_bEncOrg = bEncOrg;

  m_pDictEncoder->m_uCuIdx = cuIdx;

  m_pDictEncoder->DictCodeOneBlock();

  m_pDictEncoder->m_uPrevPackSize += m_pDictEncoder->m_uCurrPackSize;
  m_pPackSizeBuf[cuIdx] = m_pDictEncoder->m_uCurrPackSize;
}
TComDictDec::TComDictDec()
{
  m_bFirst = true;
  m_pDecBuf = NULL;
  m_uNumUnitInCu = 0;
}

void TComDictDec::InitDecoder(TComSlice* pcSlice)
{
  int frame_size = 0;
  int cuCnt;
  if (m_bFirst)
  {
    m_bFirst = false;
  }
  else
    return;

  m_uPicWidth = pcSlice->getSPS()->getPicWidthInLumaSamples();
  m_uPicHeight = pcSlice->getSPS()->getPicHeightInLumaSamples();

  m_uCuWidth = pcSlice->getSPS()->getMaxCUWidth();
  m_uCuHeight = pcSlice->getSPS()->getMaxCUHeight();

  frame_size = m_uPicWidth * m_uPicHeight;

  m_uCuLen = m_uCuWidth * m_uCuHeight * 3;
  cuCnt = pcSlice->getPic()->getPicSym()->getNumberOfCUsInFrame();

  m_uPicLenPad = cuCnt * m_uCuLen;

  m_uNumUnitInCu = (1<<22)/m_uCuLen+1;
 
  if (m_pDecBuf == NULL)
  {
    m_pDecBuf = new UChar[m_uNumUnitInCu*m_uCuLen*2*sizeof(UChar)];
    memset(m_pDecBuf,0,m_uNumUnitInCu*m_uCuLen*2*sizeof(UChar));
  }
}

void TComDictDec::destroy()
{
  if (m_pDecBuf)
  {
    delete[] m_pDecBuf;
    m_pDecBuf = NULL;
  }

}

void TComDictDec::ReOrderDataDecYuv444(UChar * pcuBuf)
{
  int i;
  UChar * pTmp;
  UChar * pY, *pU, *pV, *tY, *tU, *tV;
  pTmp = new UChar[m_uCuLen*sizeof(UChar)];
  memcpy(pTmp,pcuBuf,m_uCuLen*sizeof(UChar));
  tY = pTmp;
  tU = tY+1;
  tV = tY+2;

  pY = pcuBuf;
  pU = pY + m_uCuWidth * m_uCuHeight;
  pV = pU + m_uCuWidth * m_uCuHeight;


  for (i = 0; i< m_uCuWidth*m_uCuHeight;i++)
  {
    *pY++ = *tY;
    *pU++ = *tU;
    *pV++ = *tV;
    tY += 3;
    tU += 3;
    tV += 3;
  }

  if(pTmp)
  {
    delete[] pTmp;
    pTmp = NULL;
  }
}

void TComDictDec::TranposeCu(UChar * pCuBuf)
{
  int wdt,hgt,i,j;
  UChar *pY,*pU,*pV;
  UChar *pTmp,*tY,*tU,*tV;
  UInt cuSizeY = m_uCuWidth * m_uCuHeight;
  pTmp = new UChar[m_uCuLen*sizeof(UChar)];
  memcpy(pTmp,pCuBuf,m_uCuLen*sizeof(UChar));

  pY = pCuBuf;
  pU = pY + cuSizeY;
  pV = pU + cuSizeY;

  wdt = hgt = m_uCuWidth;
  
  for (i=0;i<wdt;i++)
  {
    tY = pTmp + i;
    for (j=0;j<hgt;j++)
    {
      *pY++ = *tY;
      tY += wdt;
    }
  }
  for (i=0;i<wdt;i++)
  {
    tU = pTmp + cuSizeY + i;
    for (j=0;j<hgt;j++)
    {
      *pU++ = *tU;
      tU += wdt;
    } 
  }
  for (i=0;i<wdt;i++)
  {
    tV = pTmp + cuSizeY*2  + i;

    for (j=0;j<hgt;j++)
    {
      *pV++ = *tV;
      tV += wdt;
    }
  }
  if (pTmp)
  {
    delete[] pTmp;
    pTmp = NULL;
  }
}

void TComDictDec::CopyCu2Dic(TComPic * pcPic, UInt cuIdx, UInt cuIdxDic, UChar * dstBase)
{
  Pel * pSrcY, *pSrcU, *pSrcV;
  UInt strideY, strideUV, i, j, cuSizeY,uvCuWidth, uvCuHeight;
  UChar * pDst, *pDstY, *pDstU, *pDstV;

  cuSizeY = m_uCuWidth * m_uCuHeight;
  
  pSrcY = pcPic->getPicYuvRec()->getAddr(COMPONENT_Y,cuIdx);
  pSrcU = pcPic->getPicYuvRec()->getAddr(COMPONENT_Cb,cuIdx);
  pSrcV = pcPic->getPicYuvRec()->getAddr(COMPONENT_Cr,cuIdx);

  pDst = dstBase + cuIdxDic * m_uCuLen;
  pDstY = pDst;
  pDstU = pDstY + cuSizeY;

  strideY = pcPic->getPicYuvRec()->getStride(COMPONENT_Y);
  
  strideUV = pcPic->getPicYuvRec()->getStride(COMPONENT_Cb);
  pDstV = pDstU + cuSizeY;
  uvCuWidth = uvCuHeight = m_uCuWidth;


  for (i=0;i<m_uCuHeight;i++)
  {
    for (j=0;j<m_uCuWidth;j++)
    {
      *pDstY++ = *pSrcY++;
    }
    pSrcY += strideY - m_uCuWidth;
  }

  for (i=0;i<uvCuHeight;i++)
  {
    for (j=0;j<uvCuWidth;j++)
    {
      *pDstU++ = *pSrcU++;
    }
    pSrcU += strideUV - uvCuWidth;
  }

  for (i=0;i<uvCuHeight;i++)
  {
    for (j=0;j<uvCuWidth;j++)
    {
      *pDstV++ = *pSrcV++;
    }
    pSrcV += strideUV - uvCuWidth;
  }
      
#if REORDER_DATA
  TranposeCu(pDst);
  ReOrderDataEncYuv444(pDst);
#endif
  
}

void TComDictDec::ReOrderDataEncYuv444(UChar * pCuBuf)
{
  int i;
  UChar * pTmp;
  UChar * pY, *tY, *tU, *tV;
  pTmp = new UChar[m_uCuLen*sizeof(UChar)];
  memcpy(pTmp,pCuBuf,m_uCuLen*sizeof(UChar));
  tY = pTmp;
  tU = tY + m_uCuWidth * m_uCuHeight;
  tV = tU + m_uCuWidth * m_uCuHeight;

  pY = pCuBuf;

  for (i = 0; i< m_uCuWidth*m_uCuHeight;i++)
  {
    *pY++ = *tY++;
    *pY++ = *tU++;
    *pY++ = *tV++;
  }

  if(pTmp)
  {
    delete[] pTmp;
    pTmp = NULL;
  }
}

void TComDictEnc::ResetForNewPic(int cuCnt)
{
  m_pDictEncoder->m_uPrevPackSize = 0;
  m_pDictEncoder->m_pStrMapBufList.clear();
  m_uiStartPackPos = 0;
}

void TComDictDec::WriteCu2Pic(TComPic* pcPic, UInt cuIdx, UInt cuIdxDic, UChar * pDecBuf) 
{ 
  int i,j, cuSize,uvCuWidth,uvCuHeight;
  int strideY, strideUV;
  Pel  *pDstY, *pDstU, *pDstV;
  UChar * pSrcY,*pSrcU,*pSrcV, *pTmp;

  cuSize = m_uCuWidth*m_uCuHeight;

  pDstY = pcPic->getPicYuvRec()->getAddr(COMPONENT_Y,cuIdx);
  pDstU = pcPic->getPicYuvRec()->getAddr(COMPONENT_Cb,cuIdx);
  pDstV = pcPic->getPicYuvRec()->getAddr(COMPONENT_Cr,cuIdx);

  strideY = pcPic->getPicYuvRec()->getStride(COMPONENT_Y);

  pSrcY = pDecBuf + cuIdxDic*m_uCuLen;

  pTmp = new UChar[m_uCuLen*sizeof(UChar)];
  memcpy(pTmp,pSrcY,m_uCuLen);

#if REORDER_DATA
  ReOrderDataDecYuv444(pTmp);
  TranposeCu(pTmp);
#endif

  pSrcY = pTmp;

  pSrcU = pSrcY + cuSize;
  pSrcV = pSrcU + cuSize;
  uvCuWidth = uvCuHeight = m_uCuWidth;
  strideUV = pcPic->getPicYuvRec()->getStride(COMPONENT_Y);

  //copy Y
  for (i=0;i<m_uCuHeight;i++)
  {
    for (j=0;j<m_uCuWidth;j++)
    {
      *pDstY++ = *pSrcY++;
    }
    pDstY += strideY - m_uCuWidth;
  } 

  //copy U
  for (i=0;i<uvCuHeight;i++)
  {
    for(j=0;j<uvCuWidth;j++)
    {
      *pDstU++ = *pSrcU++;
    }
    pDstU += strideUV - uvCuWidth;
  }

  //copy V
  for (i=0;i<uvCuHeight;i++)
  {
    for(j=0;j<uvCuWidth;j++)
    {
      *pDstV++ = *pSrcV++;
    }
    pDstV += strideUV - uvCuWidth;
  } 

  //release temp buffer
  if (pTmp)
  {
    delete[] pTmp;
    pTmp = NULL;
  }
}  
#endif
